module metric

 implicit none

 ! number of physical (not median-dual) topology 
 integer                       :: nnode  ! number of verticies
 integer                       :: nbnode ! number of verticies at body surface (exclude isothermal wall)
 integer                       :: ncell  ! number of cells
 integer                       :: nface  ! number of faces (i.e., edges in 2D)
 integer                       :: nfbnd  ! number of physical boundary faces (parallel boundary excluded)
 integer                       :: nfbnda ! number of physical & parallel boundary faces 
 integer                       :: nfbndp ! number of parallel boundary faces 
 integer                       :: nfinn  ! number of inner face (physical & parallel boundary excluded) 
 integer, allocatable          :: iface(:,:)
 integer, allocatable          :: intma(:,:)
 integer, allocatable          :: ibface(:)
 double precision, parameter   :: Pi = dacos(-1.d0)
 double precision, allocatable :: xnd(:) ! x coordinate at a vertex
 double precision, allocatable :: ynd(:) ! y coordinate at a vertex
 double precision, allocatable :: xc(:)  ! x coordinate of a cell centroid
 double precision, allocatable :: yc(:)  ! y coordinate at a cell centroid
 double precision, allocatable :: cx(:)  
 double precision, allocatable :: cy(:)  
 double precision, allocatable :: sd(:)  
 double precision, allocatable :: vc(:)  
 double precision, allocatable :: xnx(:,:)  
 double precision, allocatable :: xny(:,:)  
 double precision, allocatable :: xl(:,:)  
 double precision, allocatable :: axrc(:)    ! axisymmetric
 double precision, allocatable :: axrfb(:,:) ! axisymmetric
 double precision, allocatable :: axrf(:,:)  ! axisymmetric
 double precision, allocatable :: xf(:), xfc(:), xfcb(:,:)
 double precision, allocatable :: yf(:), yfc(:), yfcb(:,:)
 double precision, allocatable :: tsd(:,:), tcx(:,:), tcy(:,:)
end module metric


program test

#include "petsc/finclude/petscdmplex.h"
#include "petsc/finclude/petscdmlabel.h"

 use petscdmplex
 use petscsys
 use metric

 implicit none

 DM                                :: dm             ! original dm
 DM                                :: dm_geom        ! cloned dm for metrics
 DMLabel                           :: label_physical ! label for physical boundary (exclude parallel boundary)
                                                     ! Note : 'Physical' does not mean 'Solid' only. 'Physical' includes all kinds physical boundaries (e.g., inlet, outlet, solid wall, etc.)
 DMLabel                           :: label_para     ! label for parallel boundary 
 DMLabel                           :: label_all      ! label for all kinds of boundary (both physical and parallel)
 DMLabel                           :: label_marker   ! label for marker from mesh
 DMLabel                           :: label_cell     ! label for cell shape
 DMLabel                           :: label_surf_body     ! label for surface of body (e.g., spacecraft, exp material, etc.)
 DMLabel                           :: label_surf_wall_iso ! label for isothermal surface

 ! User should define this parameter
 ! ToDo : these should be loaded from Control input or hard-coded
 PetscInt, parameter               :: BC_Surf_Wall_Iso = 9  ! 1 ! what is your BC ID for "Isothermal Wall" written in mesh file?
 PetscInt, parameter               :: BC_Surf_Body     = 11 ! 2 ! what is your BC ID for "Surface Body" written in mesh file?

 Vec                               :: coord          ! coordinate array from PETSc
 Vec                               :: node           ! node looping vector
 Vec                               :: face           ! face looping vector

 PetscInt, target, dimension(4)    :: cone ! ToDo : number 4 is not general for arbitrary polygon in 2D or 3D
 PetscInt, pointer                 :: pCone(:)
 PetscInt, pointer                 :: pCone_tmp(:)


 PetscInt, target, dimension(10)   :: supp ! ToDo : number 10 is not general for arbitrary polygon in 2D or 3D
 PetscInt, pointer                 :: pSupp(:)
 


 PetscInt, parameter               :: marker_boundary  = 10 ! includes both all the physical boundaries and parallel
 PetscInt, parameter               :: marker_para      = 20 ! includes both all the physical boundaries and parallel

 PetscReal, target, dimension(2)   :: cent, norm ! ToDo : number 2 is not general for 3D. Two means x and y coordinates
 PetscReal, pointer                :: pCent(:)
 PetscReal, pointer                :: pNorm(:)

 PetscInt                          :: depth          ! depth in DAG representation
 PetscInt                          :: pst, pend      ! start & end range for points
 PetscInt                          :: vst, vend      ! start & end range for verticies
 PetscInt                          :: est, eend      ! start & end range for edges
 PetscInt                          :: cst, cend      ! start & end range for cells 
 PetscInt                          :: fst, fend      ! start & end range for faces -> not used in 2-D

 PetscReal, target, dimension(1)   :: xx, yy
 PetscReal, pointer                :: pxx(:), pyy(:)


 ! for halo communication
 PetscInt                          :: size_halo
 PetscInt                          :: remoteOffsets
 PetscSF                           :: sf, point_sf, section_sf
 PetscSection                      :: s
 type(tVec)                        :: vloc ! local vector
 type(tVec)                        :: vglo ! global vector
 ! for halo communication


 ! for PetscSection
 PetscReal                         :: one ! TMP scalar for Vec manipulation
 PetscReal                         :: two ! TMP scalar for Vec manipulation
 DMLabel, pointer                  :: nolabel(:) => NULL()
 PetscSection                      :: section
 PetscInt                          :: numFields = 3 ! Note: numFields does not mean "number of variables"
                                                    ! Example: 
                                                    ! numFields=1: put data on vertices
                                                    ! numFields=2: put data on cells
                                                    ! numFields=3: put data on faces
 PetscInt                          :: numBC     = 1
 PetscInt                          :: size
 PetscInt                          :: dim
 PetscInt, target, dimension(3)    :: numComp ! dont fix size. Its okay for Finite-Volume
 PetscInt, target, dimension(12)   :: numDof  ! dont fix size. Its okay for Finite-Volume
 PetscInt, pointer                 :: pNumComp(:)
 PetscInt, pointer                 :: pNumDof(:)
 PetscInt, target, dimension(1)    :: bcField
 PetscInt, pointer                 :: pBcField(:)
 IS, target, dimension(1)          :: bcCompIS
 IS, target, dimension(1)          :: bcPointIS
 IS, pointer                       :: pBcCompIS(:)
 IS, pointer                       :: pBcPointIS(:)

 PetscMPIInt                       :: myid, nproc, icpu ! MPI indices
 PetscBool                         :: Interpolate = PETSC_TRUE
 PetscViewer                       :: viewer
 PetscErrorCode                    :: ierr
 character(len=PETSC_MAX_PATH_LEN) :: OutStr


 ! non-PETSc but my local arrays
 integer                           :: ndim = 2          ! ToDo : non-general spatial dimension (2D:ndim=2)
 integer                           :: i, j, k, m, n, num, numa, val, nline
 integer                           :: jp, kp
 integer                           :: val_all, val_phy
 integer                           :: Iter
 integer, parameter                :: Iter_Max = 5000



 integer                           :: nd1, nd2, ndt
 integer                           :: nc1, nc2
 integer                           :: mark_phys, mark_all
 integer                           :: mark_surf, mark_para
 integer                           :: fac_para
 integer                           :: inode, iedge
 integer                           :: ifb, ifi, ifa
 integer, allocatable              :: nd(:), nd_tmp(:)
 double precision                  :: area, area_sum
 double precision, allocatable     :: sol(:)            ! tmp solution vector at a vertex
 PetscReal, target, dimension(1)   :: sol_tmp
 PetscReal, pointer                :: pSol(:)

 double precision                  :: dx2, dy2
 double precision                  :: dx3, dy3
 double precision                  :: dx4, dy4
 double precision                  :: zd, ayc
 double precision                  :: xd, yd, dot, ainv
 double precision                  :: max_er, er
 double precision                  :: sum_dci_nd1, sum_dci_nd2
 double precision                  :: dci_ocx_tmp
 double precision                  :: I_1, I_2
 double precision                  :: Qrad_1, Qrad_2
 double precision                  :: sum_isca
 double precision                  :: cxx, cyy, leng
 double precision                  :: vol_self, vol_petsc
 double precision, parameter       :: sigb=5.670400d-8, tol=1.d-6
 logical                           :: Cond_Iter 


 ! initialize petsc : This section also initializes MPI
 call PetscInitialize(PETSC_NULL_CHARACTER, ierr)
 if(ierr.ne.0) then
   write(*,*) 'Fail to initialize PETSc ...'
   stop
 endif

 ! get MPI size info 
 call MPI_COMM_RANK(PETSC_COMM_WORLD, myid, ierr)
 call MPI_COMM_SIZE(PETSC_COMM_WORLD, nproc, ierr)
 write(*,*) 'MPI information ... ', myid, nproc

 ! create mesh by default PETSc function
 call DMCreate(PETSC_COMM_WORLD, dm, ierr);CHKERRA(ierr)
 call DMSetType(dm, DMPLEX, ierr);CHKERRA(ierr)
 call DMSetFromOptions(dm, ierr);CHKERRA(ierr)
 !call DMViewFromOptions(dm, PETSC_NULL_VEC, '-dm_view', ierr);CHKERRA(ierr)
 call DMViewFromOptions(dm, PETSC_NULL_VEC, '-over_dm_view', ierr);CHKERRA(ierr)
 call DMClone(dm, dm_geom, ierr);CHKERRA(ierr)
 call PetscObjectSetName(dm_geom, 'DMPlexTest', ierr);CHKERRA(ierr)

 ! get depth of the DAG representing this mesh
 call DMPlexGetDepth(dm_geom, depth, ierr);CHKERRA(ierr)

 ! get information about spatial dimension
 call DMGetDimension(dm_geom, dim, ierr);CHKERRA(ierr)

 ! get topological information
 call DMPlexGetChart(dm_geom, pst, pend, ierr);CHKERRA(ierr) ! entire topological points (either height or depth strata)
                                                             ! including everything (i.e., cell, vertices, and interpolated topologies)

 write(OutStr,*) 'Entire topological points range ...', pst, pend, myid, '\n'
 call PetscPrintf(PETSC_COMM_SELF, OutStr, ierr);CHKERRA(ierr) 

 ! using Depth Stratum
 call DMPlexGetDepthStratum(dm_geom, 0, vst, vend, ierr);CHKERRA(ierr) ! verticies
 call DMPlexGetDepthStratum(dm_geom, 1, est, eend, ierr);CHKERRA(ierr) ! edges
 call DMPlexGetDepthStratum(dm_geom, 2, cst, cend, ierr);CHKERRA(ierr) ! cells
 nnode = vend - vst
 ncell = cend - cst
 nface = eend - est
 write(OutStr,*) 'From Depth ... cell ', cst, cend, cend-cst, myid, '\n'
 call PetscPrintf(PETSC_COMM_SELF, OutStr, ierr);CHKERRA(ierr) 
 write(OutStr,*) 'From Depth ... edge ', est, eend, eend-est, myid, '\n'
 call PetscPrintf(PETSC_COMM_SELF, OutStr, ierr);CHKERRA(ierr) 
 write(OutStr,*) 'From Depth ... vertex ', vst, vend, vend-vst, myid, '\n'
 call PetscPrintf(PETSC_COMM_SELF, OutStr, ierr);CHKERRA(ierr) 


 ! Get Physical boundary labels
 ! Note : by default, PETSc creates label named "Face Sets" when it reads a Gmsh file with some boundaries
 !        This covers all the "Physical" boundaries (e.g., inlet + outlet + ...)
 call DMGetLabel(dm_geom, "Face Sets", label_physical, ierr);CHKERRA(ierr)
 call DMPlexLabelComplete(dm_geom, label_physical, ierr);CHKERRA(ierr) ! doing this, vertex at physical boundary marked
 ! mark all kinds of boundaries (both physical and parallel)
 call DMCreateLabel(dm_geom, "All Bound", ierr);CHKERRA(ierr)
 call DMGetLabel(dm_geom, "All Bound", label_all, ierr);CHKERRA(ierr)
 call DMPlexMarkBoundaryFaces(dm_geom, marker_boundary, label_all, ierr);CHKERRA(ierr)
 call DMPlexLabelComplete(dm_geom, label_all, ierr);CHKERRA(ierr) ! doing this, vertex at physical boundary marked
 ! mark parallel boundary
 call DMCreateLabel(dm_geom, "Para Bound", ierr);CHKERRA(ierr)
 call DMGetLabel(dm_geom, "Para Bound", label_para, ierr);CHKERRA(ierr)
 do i=pst,pend
   call DMLabelGetValue(label_all, i, val_all, ierr);CHKERRA(ierr)
   call DMLabelGetValue(label_physical, i, val_phy, ierr);CHKERRA(ierr)
   if( (val_all.ne.-1) .and. (val_phy.eq.-1) ) then
     call DMLabelSetValue(label_para, i, marker_para, ierr);CHKERRA(ierr)
   endif
 enddo 
 nfbnd  = 0 ! number of boundary edge (physical)
 nfbnda = 0 ! number of boundary edge (physical & parallel)
 nfbndp = 0 ! number of boundary edge (parallel)
 do i=est, eend-1
   call DMLabelGetValue(label_physical, i, val, ierr);CHKERRA(ierr)
   if(val.ne.-1) then
     nfbnd = nfbnd + 1
   endif
   call DMLabelGetValue(label_all, i, val, ierr);CHKERRA(ierr)
   if(val.ne.-1) then
     nfbnda = nfbnda + 1
   endif
   call DMLabelGetValue(label_para, i, val, ierr);CHKERRA(ierr)
   if(val.ne.-1) then
     nfbndp = nfbndp + 1
   endif
 enddo
 nfinn = nface - nfbnda ! number of inner edge (physical & parallel boundary excluded)
 write(OutStr,*) 'Number of physical boundary edge ... ', nfbnd, myid, '\n'
 call PetscPrintf(PETSC_COMM_SELF, OutStr, ierr);CHKERRA(ierr) 
 write(OutStr,*) 'Number of physical and parallel boundary edge ... ', nfbnda, myid, '\n'
 call PetscPrintf(PETSC_COMM_SELF, OutStr, ierr);CHKERRA(ierr) 
 write(OutStr,*) 'Number of parallel boundary edge ... ', nfbndp, myid, '\n'
 call PetscPrintf(PETSC_COMM_SELF, OutStr, ierr);CHKERRA(ierr) 
 ! check
 if(nfbndp.ne.nfbnda-nfbnd) then
   write(*,*) 'Count parallel boundary edge failed ...'
   stop
 endif

 ! define section & star forest with 1-dof on vertex for halo communication
 call PetscSectionCreate(PETSC_COMM_WORLD, s, ierr);CHKERRA(ierr)
 call PetscSectionSetNumFields(s, 1, ierr);CHKERRA(ierr)
 call PetscSectionSetFieldComponents(s, 0, 1, ierr);CHKERRA(ierr)
 call PetscSectionSetChart(s, vst, vend, ierr);CHKERRA(ierr)
 do i=vst,vend-1
   call PetscSectionSetDof(s, i, 1, ierr);CHKERRA(ierr)
   call PetscSectionSetFieldDof(s, i, 0, 1, ierr);CHKERRA(ierr)
 enddo
 call PetscSectionSetUp(s, ierr);CHKERRA(ierr)
 call DMSetLocalSection(dm_geom, s, ierr);CHKERRA(ierr)


 ! halo communication between local and global vectors
 call PetscSectionDestroy(s, ierr);CHKERRA(ierr)
 call DMGetSectionSF(dm_geom, sf, ierr);CHKERRA(ierr)
 call PetscSFSetUp(sf, ierr);CHKERRA(ierr)

 ! define vectors for halo exchange 
 call DMGetLocalVector(dm_geom, vloc, ierr);CHKERRA(ierr)
 call DMGetGlobalVector(dm_geom, vglo, ierr);CHKERRA(ierr)
 call VecGetSize(vloc, size_halo, ierr);CHKERRA(ierr)


! prepare pointer for petsc vector manipulation
 sol_tmp = 0.0
 pSol => sol_tmp



 !========================================!
 ! define metric along with DMPlex layout !
 !========================================!
 ! set connectivity information through face
 allocate(iface(nface,4)) ! ToDo : number 4 is not general for 3D
 allocate(intma(ncell,4)) ! ToDo : number 4 is not general for 3D, number 4 is not general for penta & hexa mesh in 2D
 allocate(ibface(nface))  ! convert 1-nface to ifb(i.e., boundary face cursor)
                          ! used for boundary condition imposing
 iface  = -1 ! initialization
 intma  = -1 ! initialization
 ibface = -1 ! initialization

 ! iface : connectivity along face
 !   iface(i,1) : first cell index sharing given edge
 !   iface(i,2) : second cell index sharing given edge
 !   iface(i,3) : first node index sharing given edge
 !   iface(i,4) : second node index sharing given edge
 !   Note : - index i and content of iface array converted from PETSc DAG to my local
 !          - index i ranges from 1-nface, not DAG
 !          - contents of iface(i,1) & iface(i,2) range from 1-ncell, not DAG
 !          - contents of iface(i,3) & iface(i,4) range from 1-nnode, not DAG

 ! intma : connectivity along cell - ToDo: triangular & quadrilateral are okay, penta & hexa not okay
 !   intma(i,1) : first node related with given cell
 !   intma(i,2) : second node related with given cell
 !   intma(i,3) : third node related with given cell
 !   intma(i,4) : fourth node related with given cell
 !                if triangular, intma(i,4)=intma(i,3)
 !   Note : - index i and content of intma array converted from PETSc DAG to my local
 !          - index i ranges from 1-ncell, not DAG
 !          - contents of intma range from 1-nnode, not DAG

 cone  = 0
 supp  = 0
 ! cone for cell : defined by edges (e.g., for triangle : cone size=3)
 ! cone for edge : defined by vertices (e.g., cone size=2 for a given edge)
 pCone     => cone
 pCone_tmp => cone
 pSupp     => supp

 ! define label that includes cell shape information
 call DMPlexGetCellTypeLabel(dm_geom, label_cell, ierr);CHKERRA(ierr)
 do i=cst, cend-1 ! loop over cell
   ! get cone layout for given cell
   ! value of pCone implies edge that covers given cell range from {est, eend-1}
   call DMPlexGetCone(dm_geom, i, pCone, ierr);CHKERRA(ierr)
   ! get shape info for given cell
   call DMLabelGetValue(label_cell, i, nline, ierr);CHKERRA(ierr) ! get number of edges for given cell shape (2D triangle: nline=3)
   allocate(nd(nline*2))
   allocate(nd_tmp(nline))
   ! initialize
   nd     = -1
   nd_tmp = -1
   do j=1,nline ! loop over number of edge of given cell shape
     ! convert pCone(PETSc's DAG index) to local index for iface array
     iedge = pCone(j) - est + 1
     if( (iface(iedge,1).ne.-1) .and. (iface(iedge,2).ne.-1) ) then
       ! check if already filled
       write(*,*) 'redundant ... Error! ', i, j, myid
       stop
     endif
     if(iface(iedge,1) .eq. -1) then
       iface(iedge,1) = i - cst + 1
     else
       iface(iedge,2) = i - cst + 1
     endif
     ! get cone for given edge
     call DMPlexGetCone(dm_geom, pCone(j), pCone_tmp, ierr);CHKERRA(ierr)
     nd(j*2-1) = pCone_tmp(1)
     nd(j*2)   = pCone_tmp(2)
     call DMPlexRestoreCone(dm_geom, pCone(j), pCone_tmp, ierr);CHKERRA(ierr)
   enddo
   call DMPlexRestoreCone(dm_geom, i, pCone, ierr);CHKERRA(ierr)

   val = 0
   do k=1,nline*2-1   ! target 
     do m=k+1,nline*2 ! search twins (vertex)
       if(nd(m).eq.nd(k)) then 
         val = val + 1
         nd_tmp(val) = nd(k)
       endif
     enddo
   enddo
   ! check
   if(val.ne.nline) then
     write(*,*) 'node count for each cell failed ...'
     stop
   endif
   ! check 
   do k=1,nline-1
     do m=k+1,nline
       if(nd_tmp(k) .eq. nd_tmp(m)) then
         write(*,*) 'node count for each cell failed ...'
         stop
       endif
     enddo
   enddo
   ! check 
   do k=1,nline
     if(nd_tmp(k) .eq. -1) then
       write(*,*) 'node count for each cell failed ...'
       stop
     endif
   enddo
   ! assign intma array
   do k=1,nline
     intma(i-cst+1,k) = nd_tmp(k) - vst + 1
   enddo
   deallocate(nd)
   deallocate(nd_tmp)
   ! write(*,*) 'cell loop ... ', i-cst+1, (intma(i-cst+1,k),k=1,nline)
 enddo
 do i=est, eend-1 ! loop over edge
   call DMPlexGetCone(dm_geom, i, pCone, ierr);CHKERRA(ierr)
   iedge = i - est + 1
   iface(iedge,3) = pCone(1) - vst + 1 ! first node index
   iface(iedge,4) = pCone(2) - vst + 1 ! second node index
   ! write(*,*) 'edge loop ... ', i, pCone(:), myid
   call DMPlexRestoreCone(dm_geom, i, pCone, ierr);CHKERRA(ierr)
 enddo

 ! check iface info
 num  = 0
 numa = 0
 do i=1,nface ! loop over edge
   ! write(*,*) i, (iface(i,j),j=1,4)
   ! check number of physical foundary face
   do j=1,4
     if(iface(i,j).eq.-1) then             ! convert my local index to DAG
       call DMLabelGetValue(label_physical, i+est-1, val, ierr);CHKERRA(ierr)
       if(val.ne.-1) then
         num = num + 1 ! count physical boundary
       endif
       call DMLabelGetValue(label_all, i+est-1, val, ierr);CHKERRA(ierr)
       if(val.ne.-1) then
         numa = numa + 1 ! count physical & parallel boundary
       endif
     endif
   enddo
 enddo
 if(num .ne. nfbnd) then
   write(*,*) 'checking physical boundary face error in iface ...', nfbnd, num, myid
   stop
 endif
 if(numa .ne. nfbnda) then
   write(*,*) 'checking boundary face error in iface ...', nfbnda, numa, myid
   stop
 endif


 ! set vertex coordinates : extract from DMPlex
 ! get coordinate vector
 call DMGetCoordinatesLocal(dm_geom, coord, ierr);CHKERRA(ierr) ! coordinate vector for entire vertices
 call VecGetSize(coord, size, ierr);CHKERRA(ierr)
 ! size = number of local vertex * 2 (x1, y1, x2, y2, ...) in 2D
 ! size = number of local vertex * 3 (x1, y1, z1, x2, y2, z2, ...) in 3D
 allocate(xnd(nnode))
 allocate(ynd(nnode))
 ! xnd, ynd : coordinate vectors
 ! xnd(inode) : x coordinate value at inode
 ! ynd(inode) : y coordinate value at inode
 !   Note : - index inode converted from PETSc DAG to my local that ranges from 1-nnode, not DAG
 xnd = 0.d0 ; ynd = 0.d0
 inode=1       ! initialize node index
 do i=0,size-1 ! loop over vertex
   ! initialize
   xx  =  0.0 ; yy  =  0.0
   pxx => xx  ; pyy => yy ! normal PetscReal cannot be transferred to normal variable in Fortran
                          ! therefore a pointer should be used
   if(    mod(i,ndim).eq.0) then ! x coordinate
     call VecGetValues(coord, 1, i, pxx, ierr);CHKERRA(ierr)
     xnd(inode) = pxx(1) 
   elseif(mod(i,ndim).eq.1) then ! y coordinate
     call VecGetValues(coord, 1, i, pyy, ierr);CHKERRA(ierr)
     ynd(inode) = pyy(1)
   endif
   ! update node index
   if(mod(i,ndim).eq.ndim-1) then
     if(i.ne.size-1) then
       inode = inode + 1
     endif
   endif
 enddo
 ! check local node counting
 if(inode.ne.nnode) then
   write(*,*) 'node counting error in xnd ynd ...'
   stop
 endif


 ! define coordinate of cell centroid
 ! ToDo : generalize for penta and hexa -> does PETSc support this cell shape?
 allocate(xc(ncell))
 allocate(yc(ncell))
 xc = 0.d0 ; yc = 0.d0
 cent = 0.0 ; norm = 0.0
 pCent => cent
 pNorm => norm ! Note : two different pointers should be pointing two different target arrays
 do i=cst,cend-1 ! loop over cell
   call DMPlexComputeCellGeometryFVM(dm_geom, i, area, pCent, pNorm, ierr);CHKERRA(ierr)
   xc(i-cst+1) = pCent(1)
   yc(i-cst+1) = pCent(2)
 enddo

 ! define face normal vector, face area(length in 2D), median-dual volume
 allocate(xf(nfbnda), yf(nfbnda))
 allocate(xfc(nface), yfc(nface))
 allocate(xfcb(nfbnda,2), yfcb(nfbnda,2))
 allocate(tsd(nface,2), tcx(nface,2), tcy(nface,2))
 allocate(vc(nnode))
 allocate(axrc(nnode))
 allocate(axrfb(nfbnda,2), axrf(nface,2))
 allocate(cx(nfbnda), cy(nfbnda), sd(nfbnda))
 allocate(xnx(nface,2), xny(nface,2), xl(nface,2))

 ! initialize
 xf    = 0.d0 ; yf    = 0.d0
 xfc   = 0.d0 ; yfc   = 0.d0
 xfcb  = 0.d0 ; yfcb  = 0.d0
 tsd   = 0.d0 ; tcx   = 0.d0 ; tcy = 0.d0
 vc    = 0.d0
 axrc  = 0.d0
 axrfb = 0.d0 ; axrf  = 0.d0
 cx    = 0.d0 ; cy    = 0.d0 ; sd  = 0.d0
 ifb   = 0 ! boundary face cursor (physical & parallel)
 ifi   = 0 ! inner face cursor
 ifa   = 0 ! all face cursor
 do i=1,nface ! loop over entire face
   nc1 = iface(i,1)
   nc2 = iface(i,2)
   nd1 = iface(i,3)
   nd2 = iface(i,4)

   ! move face cursor
   call DMLabelGetValue(label_all, i+est-1, mark_all, ierr);CHKERRA(ierr)
   ifa = i
   if(mark_all.ne.-1) then
     ifb = ifb + 1
   else
     ifi = ifi + 1
   endif
   ibface(i) = ifb

   !===================================!
   ! physical & parallel boundary face !
   !===================================!
   if(mark_all.ne.-1) then 
   ! if( (mark_phys.ne.-1) .or. (mark_para.ne.-1) ) then 
     ! midpoint of physical boundary face
     xf(ifb) = 0.5d0*(xnd(nd1) + xnd(nd2))
     yf(ifb) = 0.5d0*(ynd(nd1) + ynd(nd2))

     ! midpoint of median-dual face, which is connected to physical boundary face
     xfc(ifa)    = 0.5d0*(xc(nc1) + xf(ifb))
     yfc(ifa)    = 0.5d0*(yc(nc1) + yf(ifb))
     axrf(ifa,1) = yfc(ifa)

     ! midpoint of median-dual boundary face
     ! close to nd1
     xfcb(ifb,1)  = 0.75d0*xnd(nd1) + 0.25d0*xnd(nd2)
     yfcb(ifb,1)  = 0.75d0*ynd(nd1) + 0.25d0*ynd(nd2)
     axrfb(ifb,1) = yfcb(ifb,1)
     ! close to nd2
     xfcb(ifb,2)  = 0.75d0*xnd(nd2) + 0.25d0*xnd(nd1)
     yfcb(ifb,2)  = 0.75d0*ynd(nd2) + 0.25d0*ynd(nd1)
     axrfb(ifb,2) = yfcb(ifb,2)

     ! vector from xf to xc
     dx2 = xc(nc1) - xf(ifb)
     dy2 = yc(nc1) - yf(ifb)

     ! vector from xf to xnd(nd1)
     dx3 = xnd(nd1) - xf(ifb)
     dy3 = ynd(nd1) - yf(ifb)

     ! vector from xf to xnd(nd2)
     dx4 = xnd(nd2) - xf(ifb)
     dy4 = ynd(nd2) - yf(ifb)

     ! length b/w midpoint of physical boundary face and cell center
     tsd(ifa,1) = dsqrt(dx2*dx2 + dy2*dy2)

     ! length b/w midpoint of physical boundary face and related node nd1
     tsd(ifa,2) = dsqrt(dx3*dx3 + dy3*dy3)

     ! area around {nd1, midpoint of boundary face, and cell center} from Gauss formula
     ! stack volume component on vertex to build median-dual control volume
     zd        = dx2*dy3 - dy2*dx3 
     area      = 0.5d0*dsqrt(zd*zd)
     vc(nd1)   = vc(nd1) + area
     ayc       = (ynd(nd1) + yf(ifb) + yc(nc1))/3.d0
     axrc(nd1) = axrc(nd1) + area    * ayc

     ! area around {nd2, midpoint of boundary face, and cell center} from Gauss formula
     ! stack volume component on vertex to build median-dual control volume
     zd        = dx2*dy4 - dy2*dx4 
     area      = 0.5d0*dsqrt(zd*zd)
     vc(nd2)   = vc(nd2) + area
     ayc       = (ynd(nd2) + yf(ifb) + yc(nc1))/3.d0
     axrc(nd2) = axrc(nd2) + area    * ayc

     ! normal vector to dx2, dy2
     xd  = -dy2
     yd  =  dx2
     dot =  dx3*xd + dy3*yd
     if(dot.gt.0.d0) then
       xd = -xd
       yd = -yd
     endif
     ! normal vector w.r.t face, which is for midpoint of boundary face and cell center
     ainv       = 1.d0/tsd(ifa,1)
     tcx(ifa,1) = xd*ainv
     tcy(ifa,1) = yd*ainv

     ! normal vector to dx3, dy3
     xd  = -dy3
     yd  =  dx3
     dot =  dx2*xd + dy2*yd
     if(dot.gt.0.d0) then
       xd = -xd
       yd = -yd
     endif
     ! normal vector w.r.t face, which is for midpoint of boundary face and related node
     ainv       = 1.d0/tsd(ifa,2)
     tcx(ifa,2) = xd*ainv
     tcy(ifa,2) = yd*ainv

     ! assign to working arrays
     ! for face b/w midpoint of boundary face and related node
     ! outgoing w.r.t nd1's & nd2's control volume (for physical boundary, outgoing for nd1's and nd2's control volume is the same)
     cx(ifb) = tcx(ifa,2)
     cy(ifb) = tcy(ifa,2)
     sd(ifb) = tsd(ifa,2)
     ! for face b/w midpoint of boundary face and cell center (part of median-dual)
     ! outgoing w.r.t nd1's control volume
     xnx(ifa,1) = tcx(ifa,1)
     xny(ifa,1) = tcy(ifa,1)
     xl(ifa,1)  = tsd(ifa,1)


   !=========================================!
   ! inner face but not in parallel boundary !
   !=========================================!
   else 
     ! midpoint of physical inner face
     xfc(ifa) = 0.5d0*(xnd(nd1) + xnd(nd2))
     yfc(ifa) = 0.5d0*(ynd(nd1) + ynd(nd2))

     ! midpoint of median-dual face for inner face
     axrf(ifa,1) = 0.5d0*(yc(nc1)+yfc(ifa)) ! for nc1 cell
     axrf(ifa,2) = 0.5d0*(yc(nc2)+yfc(ifa)) ! for nc2 cell

     !==============!
     ! for nc1 cell !
     !==============!
     ! vector from xfc to xc(nc1)
     dx2 = xc(nc1) - xfc(ifa)
     dy2 = yc(nc1) - yfc(ifa)

     ! vector from xfc to xnd(nd1)
     dx3 = xnd(nd1) - xfc(ifa)
     dy3 = ynd(nd1) - yfc(ifa)

     ! vector from xfc to xnd(nd2)
     dx4 = xnd(nd2) - xfc(ifa)
     dy4 = ynd(nd2) - yfc(ifa)

     ! length b/w midpoint of physical inner face and cell center
     tsd(ifa,1) = dsqrt(dx2*dx2 + dy2*dy2)
 
     ! area around {nd1, midpoint of inner face, and cell center} from Gauss formula
     ! stack volume component on vertex to build median-dual control volume
     zd        = dx2*dy3 - dy2*dx3 
     area      = 0.5d0*dsqrt(zd*zd)
     vc(nd1)   = vc(nd1) + area
     ayc       = (ynd(nd1) + yfc(ifa) + yc(nc1))/3.d0
     axrc(nd1) = axrc(nd1) + area     * ayc

     ! area around {nd2, midpoint of inner face, and cell center} from Gauss formula
     ! stack volume component on vertex to build median-dual control volume
     zd        = dx2*dy4 - dy2*dx4 
     area      = 0.5d0*dsqrt(zd*zd)
     vc(nd2)   = vc(nd2) + area
     ayc       = (ynd(nd2) + yfc(ifa) + yc(nc1))/3.d0
     axrc(nd2) = axrc(nd2) + area     * ayc

     ! normal vector w.r.t face, which is for midpoint of physical inner face and cell center
     xd  = -dy2
     yd  =  dx2
     dot =  dx3*xd + dy3*yd
     if(dot.gt.0.d0) then
       xd = -xd
       yd = -yd
     endif
     ainv       = 1.d0/tsd(ifa,1)
     tcx(ifa,1) = xd*ainv
     tcy(ifa,1) = yd*ainv

     ! assign to working arrays
     ! for face b/w inner face and center of nc1 cell (part of median-dual)
     ! outgoing w.r.t nd1's control volume
     xnx(ifa,1) = tcx(ifa,1)
     xny(ifa,1) = tcy(ifa,1)
     xl(ifa,1)  = tsd(ifa,1)


     !==============!
     ! for nc2 cell !
     !==============!
     ! vector from xfc to xc(nc2)
     dx2 = xc(nc2) - xfc(ifa)
     dy2 = yc(nc2) - yfc(ifa)

     ! length b/w midpoint of physical inner face and cell center
     tsd(ifa,2) = dsqrt(dx2*dx2 + dy2*dy2)
 
     ! area around {nd1, midpoint of inner face, and cell center} from Gauss formula
     ! stack volume component on vertex to build median-dual control volume
     zd        = dx2*dy3 - dy2*dx3 
     area      = 0.5d0*dsqrt(zd*zd)
     vc(nd1)   = vc(nd1) + area
     ayc       = (ynd(nd1) + yfc(ifa) + yc(nc2))/3.d0
     axrc(nd1) = axrc(nd1) + area     * ayc

     ! area around {nd2, midpoint of inner face, and cell center} from Gauss formula
     ! stack volume component on vertex to build median-dual control volume
     zd        = dx2*dy4 - dy2*dx4 
     area      = 0.5d0*dsqrt(zd*zd)
     vc(nd2)   = vc(nd2) + area
     ayc       = (ynd(nd2) + yfc(ifa) + yc(nc2))/3.d0
     axrc(nd2) = axrc(nd2) + area     * ayc

     ! normal vector w.r.t face, which is for midpoint of physical inner face and cell center
     xd  = -dy2
     yd  =  dx2
     dot =  dx3*xd + dy3*yd
     if(dot.gt.0.d0) then
       xd = -xd
       yd = -yd
     endif
     ainv       = 1.d0/tsd(ifa,2)
     tcx(ifa,2) = xd*ainv
     tcy(ifa,2) = yd*ainv

     ! assign to working arrays
     ! for face b/w inner face and center of nc2 cell (part of median-dual)
     ! outgoing w.r.t nd1's control volume
     xnx(ifa,2) = tcx(ifa,2)
     xny(ifa,2) = tcy(ifa,2)
     xl(ifa,2)  = tsd(ifa,2)

   endif
 enddo
 ! check
 if(ifb .ne. nfbnda) then
   write(*,*) 'count boundary face failed'
   stop
 endif
 if(ifi .ne. nface-nfbnda) then
   write(*,*) 'count inner face failed'
   stop
 endif


 ! check
 vol_self = 0.d0 ; vol_petsc = 0.d0
 area_sum = 0.d0
 do i=1,nnode
   area_sum = area_sum + vc(i)
 enddo
 call MPI_ALLREDUCE(area_sum, vol_self, 1, MPI_DOUBLE_PRECISION, MPI_SUM, MPI_COMM_WORLD, ierr)
 if(myid.eq.0) then 
   write(*,'(A)') 'Overall volume computed from median-dual ...'
   write(*,*) vol_self
 endif

  area_sum = 0.d0
  do i=cst,cend-1
    call DMPlexComputeCellGeometryFVM(dm_geom, i, area, pCent, pNorm, ierr);CHKERRA(ierr)
    area_sum = area_sum + area
  enddo
  call MPI_ALLREDUCE(area_sum, vol_petsc, 1, MPI_DOUBLE_PRECISION, MPI_SUM, MPI_COMM_WORLD, ierr)
  if(myid.eq.0) then 
    write(*,'(A)') 'Overall volume computed from PETSc ...'
    write(*,*) vol_petsc
  endif
  call MPI_BARRIER(MPI_COMM_WORLD, ierr)


 ! halo communication of control volume at a node
 do i=1,nnode
   pSol = vc(i) 
   call VecSetValues(vloc, 1, i-1, pSol, INSERT_VALUES, ierr);CHKERRA(ierr)
 enddo
 call VecZeroEntries(vglo, ierr);CHKERRA(ierr)
 call DMLocalToGlobalBegin(dm_geom, vloc, ADD_VALUES, vglo, ierr);CHKERRA(ierr)
 call DMLocalToGlobalEnd(dm_geom, vloc,   ADD_VALUES, vglo, ierr);CHKERRA(ierr)

 call DMGlobalToLocalBegin(dm_geom, vglo, INSERT_VALUES, vloc, ierr);CHKERRA(ierr)
 call DMGlobalToLocalEnd(dm_geom, vglo, INSERT_VALUES, vloc, ierr);CHKERRA(ierr)

 do i=1,nnode
   call VecGetValues(vloc, 1, i-1, pSol, ierr);CHKERRA(ierr)
   vc(i) = pSol(1)
 enddo
 ! halo communication of control volume at a node


 allocate(sol(nnode))
 sol = 0.d0

 !===================================!
 ! Definde section for "Node" vector !
 !===================================!
! Create a scalar field, a vector field, and a surface vector field
  numComp(:) = 1 ! Note: dont know about exact meaning yet
  pNumComp => numComp
  do i = 1, numFields*(dim+1)
     numDof(i) = 0
  end do
! define first field on vertices
  numDof(0*(dim+1)+1)     = 1 ! number of content in "vertex" field
                              ! if this number is equal to 1, scalar will be printed
                              ! if this number is lager than 1, array will be printed
! define second field on cells
  numDof(1*(dim+1)+dim+1) = 0 ! number of content in "cell" field
                              ! Note: even without defining cell field, 
                              !       Rank field still printed out
! define third field on faces
  numDof(2*(dim+1)+dim)   = 0 ! number of content in "face" field
                              ! Note: face field will not shown in VTK
  pNumDof => numDof
! Setup boundary conditions
  numBC = 1
  bcField(1) = 0
  pBcField => bcField
  call ISCreateStride(PETSC_COMM_WORLD, 1, 0, 1, bcCompIS(1), ierr);CHKERRA(ierr)
  pBcCompIS => bcCompIS
  ! Note : should turn on marker label using option (see run file)
  !        using other labels failed
  call DMGetStratumIS(dm_geom, "marker", 1, bcPointIS(1),ierr);CHKERRA(ierr)
  pBcPointIS => bcPointIS
! Create a PetscSection with this data layout
  call DMSetNumFields(dm_geom, numFields,ierr);CHKERRA(ierr)
  call DMPlexCreateSection(dm_geom,nolabel,pNumComp,pNumDof,numBC,pBcField,pBcCompIS,pBcPointIS,PETSC_NULL_IS,section,ierr)
  CHKERRA(ierr)
  call ISDestroy(bcCompIS(1), ierr);CHKERRA(ierr)
  call ISDestroy(bcPointIS(1), ierr);CHKERRA(ierr)
! Name the Field variables
  call PetscSectionSetFieldName(section, 0, 'Node', ierr);CHKERRA(ierr) ! value at node
  call DMSetLocalSection(dm_geom, section, ierr);CHKERRA(ierr)
  call PetscSectionDestroy(section, ierr);CHKERRA(ierr)

! Create a Vec with this layout of section
! Note : vector 'node' will include entire fields defined above. 
  !call DMGetGlobalVector(dm_geom, node, ierr);CHKERRA(ierr) 
  call DMGetLocalVector(dm_geom, node, ierr);CHKERRA(ierr) 

  ! Note : Local vector size depends on the way section defined
  call VecGetSize(node, size, ierr);CHKERRA(ierr)

  call PetscObjectSetName(node, 'ControlVolume', ierr);CHKERRA(ierr)

 do i=0,size-1 ! vertex first and then face or vertex : according to your section definition
   ! example - 2 : pass arbitrary non-petsc solution vector to petsc data layout
   pSol = vc(i+1)
   !pSol = sol(i+1)
   call VecSetValues(node, 1, i, pSol, INSERT_VALUES, ierr);CHKERRA(ierr)
 enddo


 ! create viewer
 call PetscViewerCreate(PETSC_COMM_WORLD, viewer, ierr);CHKERRA(ierr)

 ! set viewer type
 call PetscViewerSetType(viewer, PETSCVIEWERVTK, ierr);CHKERRA(ierr)

 ! set output file name for viewer
 call PetscViewerFileSetName(viewer, 'sol.vtu', ierr);CHKERRA(ierr) 

 ! print dm object to the file
 call VecView(node, viewer, ierr);CHKERRA(ierr)


 ! destroy
 call PetscViewerDestroy(viewer, ierr);CHKERRA(ierr)
 !call DMRestoreGlobalVector(dm_geom, node, ierr);CHKERRA(ierr)
 call DMRestoreLocalVector(dm_geom, node, ierr);CHKERRA(ierr)
 call DMDestroy(dm, ierr);CHKERRA(ierr)
 call DMDestroy(dm_geom, ierr);CHKERRA(ierr)
 ! deallocate non-petsc arrays
 deallocate(iface, intma)
 deallocate(xnd, ynd)
 deallocate(xc, yc)
 deallocate(sol)
 deallocate(xf, xfc, xfcb)
 deallocate(yf, yfc, yfcb)
 deallocate(tsd, tcx, tcy)

 ! finalize petsc
 write(OutStr,*) 'done ...', myid, '\n'
 call PetscPrintf(PETSC_COMM_SELF, OutStr, ierr);CHKERRA(ierr) 
 call PetscFinalize(ierr)

 stop
end program test
